From 31288b90d074681058a0eeec0d6bea66f14bcaf3 Mon Sep 17 00:00:00 2001
From: "houzj.fnst" <houzj.fnst@cn.fujitsu.com>
Date: Thu, 2 Dec 2021 15:44:19 +0800
Subject: [PATCH] Fix double publish of child table's data.

We publish the child table data twice for publication that has both
child and parent tables and is published with publish_via_partition_root
as true. This happens because subscribers will initiate synchronization
using both parent and child table as it gets both as separate tables
in the initial table list.

Ensure that pg_publication_tables returns only parent tables in such
cases.

Author: Hou Zhijie
Reviewed-by: Amit Kapila, Greg Nancarrow
Discussion: https://postgr.es/m/OS0PR01MB57167F45D481F78CDC5986F794B99@OS0PR01MB5716.jpnprd01.prod.outlook.com
Backpatch-through: 13
---
 doc/src/sgml/catalogs.sgml           | 10 +++++++++
 src/backend/catalog/pg_publication.c | 41 ++++++++++++------------------------
 2 files changed, 24 insertions(+), 27 deletions(-)

diff --git a/doc/src/sgml/catalogs.sgml b/doc/src/sgml/catalogs.sgml
index c1d11be..b2134a9 100644
--- a/doc/src/sgml/catalogs.sgml
+++ b/doc/src/sgml/catalogs.sgml
@@ -11350,6 +11350,16 @@ SELECT * FROM pg_locks pl LEFT JOIN pg_prepared_xacts ppx
    there will be a row for each eligible table.
   </para>
 
+  <para>
+   For publications of partitioned tables with
+   <literal>publish_via_partition_root</literal> set to
+   <literal>true</literal>, only the partitioned table (and not its partitions)
+   is included in the view, whereas if
+   <literal>publish_via_partition_root</literal> is set to
+   <literal>false</literal>, only the individual partitions are included in the
+   view.
+  </para>
+
   <table>
    <title><structname>pg_publication_tables</structname> Columns</title>
    <tgroup cols="1">
diff --git a/src/backend/catalog/pg_publication.c b/src/backend/catalog/pg_publication.c
index 63579b2..96826e1 100644
--- a/src/backend/catalog/pg_publication.c
+++ b/src/backend/catalog/pg_publication.c
@@ -142,7 +142,7 @@ is_publishable_class(Oid relid, Form_pg_class reltuple)
  * the publication.
  */
 static List *
-filter_partitions(List *relids, List *schemarelids)
+filter_partitions(List *relids)
 {
 	List	   *result = NIL;
 	ListCell   *lc;
@@ -161,16 +161,8 @@ filter_partitions(List *relids, List *schemarelids)
 		{
 			Oid			ancestor = lfirst_oid(lc2);
 
-			/*
-			 * Check if the parent table exists in the published table list.
-			 *
-			 * XXX As of now, we do this if the partition relation or the
-			 * partition relation's ancestor is present in schema publication
-			 * relations.
-			 */
-			if (list_member_oid(relids, ancestor) &&
-				(list_member_oid(schemarelids, relid) ||
-				 list_member_oid(schemarelids, ancestor)))
+			/* Check if the parent table exists in the published table list. */
+			if (list_member_oid(relids, ancestor))
 			{
 				skip = true;
 				break;
@@ -879,22 +871,17 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 															PUBLICATION_PART_ROOT :
 															PUBLICATION_PART_LEAF);
 			tables = list_concat_unique_oid(relids, schemarelids);
-			if (schemarelids && publication->pubviaroot)
-			{
-				/*
-				 * If the publication publishes partition changes via their
-				 * respective root partitioned tables, we must exclude
-				 * partitions in favor of including the root partitioned
-				 * tables. Otherwise, the function could return both the child
-				 * and parent tables which could cause data of the child table
-				 * to be double-published on the subscriber side.
-				 *
-				 * XXX As of now, we do this when a publication has associated
-				 * schema or for all tables publication. See
-				 * GetAllTablesPublicationRelations().
-				 */
-				tables = filter_partitions(tables, schemarelids);
-			}
+
+			/*
+			 * If the publication publishes partition changes via their
+			 * respective root partitioned tables, we must exclude partitions
+			 * in favor of including the root partitioned tables. Otherwise,
+			 * the function could return both the child and parent tables
+			 * which could cause data of the child table to be
+			 * double-published on the subscriber side.
+			 */
+			if (publication->pubviaroot)
+				tables = filter_partitions(tables);
 		}
 
 		funcctx->user_fctx = (void *) tables;
-- 
2.7.2.windows.1

